
import unittest
import numpy as np
import numpy.testing as npt
from itertools import product

from mogu.tensor import decompose_tensor_jennrich, decompose_tensor_ALS


class test_tensordecomposition(unittest.TestCase):
    def test_jennrich(self):
        # definining a rank-3 tensor
        x1 = np.array([[-1.3, 1], [2, -1.4]], dtype=np.float32)
        x2 = np.array([[1, 0], [-1.5, 0.6]], dtype=np.float32)
        x = np.zeros((2, 2, 2))
        x[:, :, 0] = x1
        x[:, :, 1] = x2

        # Jennrich decomposition
        eigvecs_u, eigvecs_v, eigvecs_w = decompose_tensor_jennrich(x)

        # verification
        xh = np.zeros(x.shape)
        for i, j, k in product(range(x.shape[0]), range(x.shape[1]), range(x.shape[2])):
            xh[i, j, k] = sum([eigvecs_u[i, alpha] * eigvecs_v[j, alpha] * eigvecs_w[k, alpha] for alpha in
                               range(eigvecs_u.shape[1])])

        # asserting
        npt.assert_almost_equal(x, xh, decimal=4)

    def test_als(self):
        # initialize matrix
        x1 = np.array([[1, 0], [0, 1]], dtype=np.float32)
        x2 = np.array([[0, -1], [1, 0]], dtype=np.float32)
        x = np.zeros((2, 2, 2))
        x[:, :, 0] = x1
        x[:, :, 1] = x2

        A, B, C = decompose_tensor_ALS(x, k=2, nbiter=10000)

        # validation
        mat = np.zeros((A.shape[0], B.shape[0], C.shape[0]))
        for i, j, k in product(range(A.shape[0]), range(B.shape[0]), range(C.shape[0])):
            mat[i, j, k] = sum(A[i, alpha] * B[j, alpha] * C[k, alpha] for alpha in range(2))

        # npt.assert_almost_equal(x, mat, decimal=1)  # cannot really make this work
        assert True



if __name__ == '__main__':
    unittest.main()



import numpy as np


def FFT(f):
    nbdata = f.shape[0]
    assert np.log(nbdata) / np.log(2) % 1 == 0.0

    if nbdata == 1:
        return f

    x = FFT(f[0::2])
    y = FFT(f[1::2])

    g = np.zeros(nbdata, dtype=np.complex_)
    for i in range(nbdata // 2):
        w = np.exp(-1j * 2 * np.pi * i / nbdata)
        g[i] = x[i] + w * y[i]
        g[i + nbdata // 2] = x[i] - w * y[i]

    return g


def invFFTrecursive(f):
    nbdata = f.shape[0]
    assert np.log(nbdata) / np.log(2) % 1 == 0.0

    if nbdata == 1:
        return f

    x = invFFTrecursive(f[0::2])
    y = invFFTrecursive(f[1::2])

    g = np.zeros(nbdata, dtype=np.complex_)
    for i in range(nbdata // 2):
        w = np.exp(1j * 2 * np.pi * i / nbdata)
        g[i] = x[i] + w * y[i]
        g[i + nbdata // 2] = x[i] - w * y[i]

    return g


def invFFT(f):
    nbdata = f.shape[0]
    return invFFTrecursive(f) / nbdata





import numpy as np
from itertools import product
from tfnumpy.tensor import rank3tensor_decomposition_ALS

def decompose_tensor_jennrich(rank3tensor):
    # initialize two random variables
    a = np.random.uniform(size=rank3tensor.shape[2])
    b = np.random.uniform(size=rank3tensor.shape[2])

    # T_a and T_b
    Ta = sum(rank3tensor[:, :, i ] *a[i] for i in range(rank3tensor.shape[2]))
    Tb = sum(rank3tensor[:, :, i ] *b[i] for i in range(rank3tensor.shape[2]))

    # eigenvalues of various auxilliary matrices
    eigvals_u, eigvecs_u = np.linalg.eig(np.matmul(Ta, np.linalg.pinv(Tb)))
    eigvals_v, eigvecs_v = np.linalg.eig(np.transpose(np.matmul(np.linalg.pinv(Ta), Tb)))

    # pair up reciprocal eigenvalues
    # pair up eigenvalues of Ta and Tb
    idx_pairs = []
    tol = 1e-5

    for i, eigval_u in enumerate(eigvals_u):
        for j, eigval_v in enumerate(eigvals_v):
            if abs(eigval_u - 1/ eigval_v) < tol:
                idx_pairs += [(i, j)]
                break

    # solving for third eigenvectors
    nbcomp = len(idx_pairs)
    solved = False
    while not solved:
        try:
            A = np.zeros((nbcomp * rank3tensor.shape[2], nbcomp * rank3tensor.shape[2]))
            B = np.zeros(nbcomp * rank3tensor.shape[2])
            eqidx = 0
            ij_combs = list(tuple(product(range(rank3tensor.shape[0]), range(rank3tensor.shape[1]))))
            for k in range(rank3tensor.shape[2]):
                for ij_comb_idx in np.random.choice(range(len(ij_combs)), size=nbcomp, replace=False):
                    i, j = ij_combs[ij_comb_idx]
                    B[eqidx] = rank3tensor[i, j, k]
                    for ck in range(nbcomp):
                        A[eqidx, ck * rank3tensor.shape[2] + k] = eigvecs_u[i, idx_pairs[ck][0]] * eigvecs_v[j, idx_pairs[ck][1]]
                    eqidx += 1

            sol = np.linalg.solve(A, B)
            solved = True   # exception is not caught at this point
        except np.linalg.LinAlgError:
            solved = False
    eigvecs_w = sol.reshape((nbcomp, rank3tensor.shape[2]), order='F')

    # rearranging eigenvectors
    rearranged_eigvecs_u = np.zeros(shape=eigvecs_u.shape)
    rearranged_eigvecs_v = np.zeros(shape=eigvecs_v.shape)
    for i, (u_idx, v_idx) in enumerate(idx_pairs):
        rearranged_eigvecs_u[:, i] = eigvecs_u[:, u_idx]
        rearranged_eigvecs_v[:, i] = eigvecs_v[:, v_idx]
    rearranged_eigvecs_w = eigvecs_w

    # return values
    return rearranged_eigvecs_u, rearranged_eigvecs_v, rearranged_eigvecs_w


def decompose_tensor_ALS(rank3tensor, k, *args, **kwargs):
    return rank3tensor_decomposition_ALS(rank3tensor, k, *args, **kwargs)




import unittest

import numpy as np

from mogu.spectral import slowDFT, slowInvDFT, FFT, invFFT

class test_finance(unittest.TestCase):
    def test_dft(self):
        x = np.array([0, -1.5, -3., -1.5, 0, 1.5, 3., 1.5, 0.])
        k = slowDFT(x)
        inv_x = slowInvDFT(k)
        np.testing.assert_array_almost_equal(x, np.real(inv_x))

    def test_fft(self):
        x = np.array([.25, .75, .75, .25, -.25, -.75, -.75, -.25])
        k = FFT(x)
        inv_x = invFFT(k)
        np.testing.assert_array_almost_equal(x, np.real(inv_x))


if __name__ == '__main__':
    unittest.main()




import numpy as np


def slowDFT(t):
    N = t.shape[0]
    omega = np.zeros(N, dtype=np.complex_)
    for n in range(N):
        omega[n] = np.sum(t[m] * np.exp(-1j*2*np.pi*m*n/N) for m in range(N))
    return omega


def slowInvDFT(omega):
    N = omega.shape[0]
    t = np.zeros(N, dtype=np.complex_)
    for m in range(N):
        t[m] = np.sum(omega[n] * np.exp(1j*2*np.pi*m*n/N) for n in range(N)) / N
    return t
